#!/usr/bin/perl -w
use strict;
use Getopt::Long;
use File::Basename;
use List::Util qw[min max];


###############################################################
##################### DOCUMENTATION ###########################

## Author: Yun Li
## Date: 2010-12-06
## Goal: Split reference haplotype files into smaller pieces
my $location = $0;
my $title = "splitRef version 0.0.2";

print "$title\n";
print "\t\@: $location\n";
print "\tby: Yun Li(ylwtx\@med.unc.edu)\n\n";

## changelog
## 0.0.2
##      fix a bug when nWindows specified
##      Use of uninitialized value within @haparr in string at /home/ylwtx/codes/pl/splitRef/splitRef_V001/splitRef.pl line 551, <IN> line ..
##      fantasia:~/1000Genomes/2010-11/BI/BI_inMerged2of4intersection_20100804> 
##         ~/codes/pl/splitRef/splitRef_V001/splitRef.pl -hap hap/EUR/EUR.chr22.hap.gz -snps snps/EUR.chr22.snps -nWindows 2 -overlapSize 500000 -o split


## Input Parameters
##

## Input Files
##

## Output Files

################# END OF DOCUMENTATION ########################
###############################################################
my $starttime = time();

my $MIN_NUM_OPTS = 4;

if($#ARGV<($MIN_NUM_OPTS*2 - 1))
{
	&usage();
	die "Failed to parse parameters\n\n";
}


my %opts = ();

	# Default Optional Options
$opts{extimateWindowOnly} = 0;
$opts{o} = "splitRef.out";

Getopt::Long::GetOptions(\%opts,qw(
	nWindows=i
	windowSize=i
	windowLength=i
	overlapSize=i
	overlapLength=i
	extimateWindowOnly=i
	hap=s
	snps=s
	map=s
	o=s
)) || die "Failed to parse options\n\n";


  ###################
  ## sanity checks ##
  ###################
if ( (!defined $opts{hap}) || (!defined $opts{snps}))
{
	&usage();
	die "Failed to parse parameters\n\n";
}
if ( (!defined $opts{nWindows}) && (!defined $opts{windowSize}) && (!defined $opts{windowLength}))
{
  print "ERROR: must specify window size by specifying either (1) number of windows -nWindows; or (2) number of markers in window -windowSize; or (3) window length (in bp) -windowLength\n\n";
	&usage();
	die "Failed to parse parameters\n\n";
}
if ( (!defined $opts{overlapSize}) && (!defined $opts{overlapLength}))
{
  print "ERROR: must specify overlapping window size by specifying either (1) number of flanking markers -overlapSize; or (2) length of a flanking region (in bp) -overlapLength\n\n";
	&usage();
	die "Failed to parse parameters\n\n";
}

&printPar();

my $nmarkers; # total number of records in .snps file;
my $totalLength;
&main();

my $endtime = time();
my $lapsetime = $endtime - $starttime;
print "\nAnalysis took $lapsetime seconds\n\n";

# begin sub

sub usage
{
	print "\n";
	print "Usage:\n\t";
	print "-hap\t Input haplotype file (Required) \n\t";
	print "-snps\t Input snp list file (Required) \n\t";
	print "-map\t Input map file (Optional) \n\t";
	print "-nWindows\t # windows to split into (Co-Required Set 1) \n\t";
	print "-windowSize\t window size in terms of # of markers per window (Co-Required Set 1) \n\t";
	print "-windowLength\t window size in terms of # of base pairs (Co-Required Set 1, and can only be used when map file specified) \n\t";
	print "-overlapSize\t # markers overlapping between two consecutive windows (Co-Required Set 2) \n\t";
	print "-overlapLength\t # base pairs overlapping between two consecutive windows (Co-Required Set 2) \n\t";
	print "-extimateWindowOnly\t Only estimate # windows, window size/length, and overlapping size/length (Optional, default=0) \n\t";
	print "-o\t Output prefix (Optional, default=splitRef.out) \n\t";
	print "\n";
}


sub printPar
{
	print "\n";
	print "Parameters in Effect:\n";
	print "\t Input haplotype file \n\t\t-hap '$opts{hap}'\n";
	print "\t Input snp list file \n\t\t-snps '$opts{snps}'\n";
	if (defined $opts{map}) {print "\t Input map file \n\t\t-map '$opts{map}'\n";}
	if (defined $opts{nWindows}) {print "\t # windows to split into \n\t\t-nWindows '$opts{nWindows}'\n";}
	if (defined $opts{windowSize}) {print "\t window size in terms of # of markers per window \n\t\t-windowSize '$opts{windowSize}'\n";}
	if (defined $opts{windowLength}) {print "\t window size in terms of # of base pairs \n\t\t-windowLength '$opts{windowLength}'\n";}
	if (defined $opts{overlapSize}) {print "\t # markers overlapping between two consecutive windows \n\t\t-overlapSize '$opts{overlapSize}'\n";}
	if (defined $opts{overlapLength}) 
	{
  	print "\t # base pairs overlapping between two consecutive windows \n\t\t-overlapLength '$opts{overlapLength}'\n";
  }
	print "\t Only estimate # windows, window size/length, and overlapping size/length \n\t\t-extimateWindowOnly '$opts{extimateWindowOnly}'\n";
	print "\t Output prefix \n\t\t-o '$opts{o}'\n";
	print "\n";
}

sub rm
{
	if(-e $_[0])
	{
		print "WARNING: $_[0] existed and removed!!\n\n";
		system "/bin/rm -rf $_[0]";
	}
}

#==================================================================
# Subroutine:
#   Open($filename, $handle)
#
# Usage:
#   Open('myfile.txt', \*IN);
#   Open($pedfile, \*PED);
#
#   Then refernce the handles <IN> and <PED> as usual.
#
# Description:
#   Open the file 'filename'. The file can be a normal file
#   or one compressed with gzip or bzip2.
#   Function does not return if the open fails.
#
# Arguments:
#   $filename - path to file to open
#   $handle - GLOB name of handle to use. Include '\*'
#
#==================================================================
sub Open {
    my ($filename, $handle) = @_;

    # Handle gzipped data
    if ($filename =~ /\.gz$/) {
        my $cmd = "gunzip -c $filename";
        open($handle, "$cmd |") ||
            die "Unable to invoke command '$cmd': $!\n";
        if (eof($handle)) { die "Cmd failed: '$cmd'\n"; }
        return 1;
    }
    # Handle bzip2 data
    if ($filename =~ /\.bz2$/) {
        my $cmd = "bunzip2 -c $filename";
        open($handle, "$cmd |") ||
            die "Unable to invoke command '$cmd': $!\n";
        if (eof($handle)) { die "Cmd failed: '$cmd'\n"; }
        return 2;
    }
    # Handle normal file
    open($handle, $filename) ||
        die "Unable to open file '$filename': $!\n";
    return 3;
}


sub main
{
  ###################
  ## sanity checks ##
  ###################
  if (!(defined $opts{map}))
  {
    if ( (defined $opts{overlapLength}) || (defined $opts{windowLength}))
    {
      print "ERROR: neither overlapLength nor windowLength is allowed because map file not specified.\n\n";
      exit;
    }
  }
  
  ##########################
  ## estimate window size ##
  ##########################
  &estimateWindow();
  
  if ($opts{extimateWindowOnly} ne 0)
  {
    exit;
  }
  
  ###############################
  ## split the reference files ##
  ###############################
  &split();
  
}

sub estimateWindow
{
  #########################
  ## part 1: core window ##
  #########################
  my $coreMarkers;
  my $nWindows;
  my $endPos;
  my $startPos;
  
  $nmarkers = `wc -l < $opts{snps}`;
  chomp($nmarkers);
  if (defined $opts{overlapSize})
  {
    if ($opts{overlapSize} >= $nmarkers)
    {
      print "WARNING: #markers overlapping specified $opts{overlapSize} > total number of markers ($nmarkers). No splitting will be performed.\n\n";
      exit;
    }
  }
  ######################################
  ## scenario 1.1: nWindows specified ##
  ######################################
  if (defined $opts{nWindows})
  {
    if ( (defined $opts{windowSize}) || (defined $opts{windowLength}) )
    {
      print "ERROR: Specify one and only one of the following three: nWindows, windowSize, windowLength\n\n";
      exit;
    }
    $nWindows = $opts{nWindows};
    $coreMarkers = int($nmarkers / $opts{nWindows});
  }
  
  ########################################
  ## scenario 1.2: windowSize specified ##
  ########################################
  if (defined $opts{windowSize})
  {
    if ( (defined $opts{nWindows}) || (defined $opts{windowLength}))
    {
      print "ERROR: Specify one and only one of the following three: nWindows, windowSize, windowLength\n\n";
      exit;
    }
    $coreMarkers = $opts{windowSize};
    $nWindows = $nmarkers / $opts{windowSize};
    if ( ($nWindows - int($nWindows)) > 0.5) {$nWindows = int($nWindows) + 1;}
    else {$nWindows = int($nWindows); }
  }
  
  ##########################################
  ## scenario 1.3: windowLength specified ##
  ##########################################
  if (defined $opts{map})
  {
    $endPos = `tail -n 1 $opts{map} | awk '{print \$3}'`;
    chomp($endPos);
    $startPos = `head -1 $opts{map} | awk '{print \$3}'`;
    chomp($startPos);
    $totalLength = $endPos + 1 - $startPos;
  }
  if (defined $opts{windowLength})
  {
    if ( (defined $opts{windowSize}) || (defined $opts{nWindows}) )
    {
      print "ERROR: Specify one and only one of the following three: nWindows, windowSize, windowLength\n\n";
      exit;
    }
    $nWindows = $totalLength / $opts{windowLength};
    if ( ($nWindows - int($nWindows)) > 0.5) {$nWindows = int($nWindows) + 1;}
    else {$nWindows = int($nWindows); }
  }
  
  #############################
  ## part 2: flanking window ##
  #############################
  my $flankingMarkers;
  #########################################
  ## scenario 2.1: overlapSize specified ##
  #########################################
  if (defined $opts{overlapSize})
  {
    if (defined $opts{overlapLength})
    {
      print "ERROR: Specify one and only one of the following two: overlapSize, overlapLength\n\n";
      exit;
    }
    $flankingMarkers = $opts{overlapSize};
  }
  
  ###########################################
  ## scenario 2.2: overlapLength specified ##
  ###########################################
  if (defined $opts{overlapLength})
  {
    if (defined $opts{overlapSize})
    {
      print "ERROR: Specify one and only one of the following two: overlapSize, overlapLength\n\n";
      exit;
    }
  }

  ##################################
  ## part 3: print the statistics ##
  ##################################
  if ( (!(defined $opts{overlapLength})) && (!(defined $opts{windowLength})) )
  {
    print "\nWill split total $nmarkers SNPs into $nWindows windows, each with [$coreMarkers + 2 * $flankingMarkers ] markers in window\n\n";
  }
  else
  {
    print "\nWill split total $totalLength base pairs into the following $nWindows windows:\n";
    print "window [windowStart (coreStart coreEnd) windowEnd] #leftFlankingSNPs #coreSNPs #rightFlankingSNPs #totalSNPs\n";
    if ( (defined $opts{overlapLength}) && (defined $opts{windowLength}) )
    {
      for (my $i = 0; $i < $nWindows; $i ++)
      {
        my $windowIndex = $i + 1;
        my $coreStart = $opts{windowLength} * $i + $startPos;
        my $coreEnd = $coreStart + $opts{windowLength};
        my $coreSNPs = `awk -v s=$coreStart -v e=$coreEnd '\$3 >= s && \$3 <= e' $opts{map} | wc -l`;
        chomp($coreSNPs);
        
        # the following would work for the first and last region;
        my $winStart = $coreStart - $opts{overlapLength};
        my $winEnd = $coreEnd + $opts{overlapLength};
        if ($windowIndex == $nWindows) {$winEnd = $endPos;}
        my $winSNPs = `awk -v s=$winStart -v e=$winEnd '\$3 >= s && \$3 <= e' $opts{map} | wc -l`;
        chomp($winSNPs);
        
        # check if left overlapping region contains any markers
        my $leftSNPs = `awk -v s=$winStart -v e=$coreStart '\$3 >= s && \$3 < e' $opts{map} | wc -l`;
        chomp($leftSNPs);
        
        # check if right overlapping region contains any markers
        my $rightSNPs = `awk -v s=$coreEnd -v e=$winEnd '\$3 > s && \$3 <= e' $opts{map} | wc -l`;
        chomp($rightSNPs);
        
        print "$windowIndex [$winStart ($coreStart $coreEnd) $winEnd] $leftSNPs $coreSNPs $rightSNPs $winSNPs\n";
        ## sanity checks 
        if ($leftSNPs == 0 && $windowIndex != 1)
        {
          print "ERROR: NO markers in the left flanking region for window $windowIndex\n";
          exit;
        }
        if ($rightSNPs == 0 && $windowIndex != $nWindows)
        {
          print "ERROR: NO markers in the right flanking region for window $windowIndex\n";
          exit;
        }
      }
    }
    else
    {
      print "ERROR: only one of overlapLength and windowLength is specified. Specify either both or none\n\n";
      exit;
    }
  }
}


sub split
{
  #########################
  ## part 1: core window ##
  #########################
  my $coreMarkers;
  my $nWindows;
  my $endPos;
  my $startPos;
  
  $nmarkers = `wc -l < $opts{snps}`;
  chomp($nmarkers);
  ######################################
  ## scenario 1.1: nWindows specified ##
  ######################################
  if (defined $opts{nWindows})
  {
    if ((defined $opts{windowSize}) || (defined $opts{windowLength}) )
    {
      print "ERROR: Specify one and only one of the following three: nWindows, windowSize, windowLength\n\n";
      exit;
    }
    $nWindows = $opts{nWindows};
    $coreMarkers = int($nmarkers / $opts{nWindows});
  }
  
  ########################################
  ## scenario 1.2: windowSize specified ##
  ########################################
  if (defined $opts{windowSize})
  {
    if ( (defined $opts{nWindows}) || (defined $opts{windowLength}) )
    {
      print "ERROR: Specify one and only one of the following three: nWindows, windowSize, windowLength\n\n";
      exit;
    }
    $coreMarkers = $opts{windowSize};
    $nWindows = $nmarkers / $opts{windowSize};
    if ( ($nWindows - int($nWindows)) > 0.5) {$nWindows = int($nWindows) + 1;}
    else {$nWindows = int($nWindows); }
  }
  
  ##########################################
  ## scenario 1.3: windowLength specified ##
  ##########################################
  if (defined $opts{map})
  {
    $endPos = `tail -n 1 $opts{map} | awk '{print \$3}'`;
    chomp($endPos);
    $startPos = `head -1 $opts{map} | awk '{print \$3}'`;
    chomp($startPos);
    $totalLength = $endPos + 1 - $startPos;
  }
  if (defined $opts{windowLength})
  {
    if ( (defined $opts{windowSize}) || (defined $opts{nWindows}) )
    {
      print "ERROR: Specify one and only one of the following three: nWindows, windowSize, windowLength\n\n";
      exit;
    }
    $nWindows = $totalLength / $opts{windowLength};
    if ( ($nWindows - int($nWindows)) > 0.5) {$nWindows = int($nWindows) + 1;}
    else {$nWindows = int($nWindows); }
  }
  
  #############################
  ## part 2: flanking window ##
  #############################
  my $flankingMarkers;
  #########################################
  ## scenario 2.1: overlapSize specified ##
  #########################################
  if (defined $opts{overlapSize})
  {
    if (defined $opts{overlapLength})
    {
      print "ERROR: Specify one and only one of the following two: overlapSize, overlapLength\n\n";
      exit;
    }
    $flankingMarkers = $opts{overlapSize};
  }
  
  ###########################################
  ## scenario 2.2: overlapLength specified ##
  ###########################################
  if (defined $opts{overlapLength})
  {
    if (defined $opts{overlapSize})
    {
      print "ERROR: Specify one and only one of the following two: overlapSize, overlapLength\n\n";
      exit;
    }
  }

  ###################
  ## part 3: split ##
  ###################
  for (my $i = 0; $i < $nWindows; $i ++)
  {
    my $index = $i + 1;
    
    #############################
    ## generate output file(s) ##
    #############################
      ## create output handlers ##
      my $osf = $opts{o}.".".$index.".snps";
      my $ohf = $opts{o}.".".$index.".hap";
      &rm($ohf);
      open(OHF, ">>$ohf") || die "cannot create file $ohf\n\n";
      
      ## determine marker index of starting and ending SNPs in window ##
      my $startIndex;
      my $endIndex;
    
      if ( (!(defined $opts{overlapLength})) && (!(defined $opts{windowLength})) )
      {
        $startIndex = $i * $coreMarkers - $flankingMarkers;
        if ($i == 0) {$startIndex += $flankingMarkers;}
        $endIndex = $index * $coreMarkers + $flankingMarkers;
        if ($index == $nWindows) {$endIndex = $nmarkers - 1;}
      }
      else
      {
        if ( (defined $opts{overlapLength}) && (defined $opts{windowLength}) )
        {
          my $coreStart = $opts{windowLength} * $i + $startPos;
          my $coreEnd = $coreStart + $opts{windowLength};
          
          # the following would work for the first and last region;
          my $winStart = $coreStart - $opts{overlapLength};
          my $winEnd = $coreEnd + $opts{overlapLength};
          if ($index == $nWindows) {$winEnd = $endPos;}
          
          $startIndex = `cat -n $opts{map} | awk -v s=$winStart -v e=$winEnd '\$4 >= s && \$4 <= e' | awk '{print \$1}' | head -1` - 1;
          $endIndex = `cat -n $opts{map} | awk -v s=$winStart -v e=$winEnd '\$4 >= s && \$4 <= e' | awk '{print \$1}' | tail -n 1` - 1;
          
          #print "$startIndex $endIndex\n";
        }
        else
        {
          print "ERROR: only one of overlapLength and windowLength is specified. Specify either both or none\n\n";
          exit;
        }
      }

      if($startIndex < 0) {$startIndex = 0;}
      if($endIndex >= $nmarkers) {$endIndex = $nmarkers - 1;}
      ## generate output ##
      print "Generating output for window $index ...\n";
        ## generate snps file ##
        my $startLine = $startIndex + 1;
        my $endLine = $endIndex + 1;
        my $nLines = $endIndex - $startIndex + 1;
        system "head -$endLine $opts{snps} | tail -n $nLines > $osf";
        
        ## generate hap file ##
        Open($opts{hap},\*IN) || die "cannot open file $opts{hap}\n\n";
        my $linenum = 0;
        while(my $line = <IN>)
        {
          $linenum ++;
          chomp($line);
          my @linearr = (split(/\s+/,$line));
          my @haparr  = (split(//,$linearr[$#linearr])); # last field contains the alleles
          my $obsfields = scalar(@haparr);
          if ($obsfields != $nmarkers)
          {
            print "ERROR: line $linenum of opts{hap} contains $obsfields alleles, while expecting $nmarkers based on $opts{snps}\n\n";
            exit;
          }
          
          # print out heading columns, if existing
          for (my $k = 0; $k < $#linearr; $k ++)
          {
            print OHF "$linearr[$k] ";
          }
          
          # print out the alleles
          for (my $k = $startIndex; $k <= $endIndex; $k ++)
          {
            print OHF "$haparr[$k]";
          }
          print OHF "\n";
        }
        close(IN);
 
      close(OHF);
    
    #############################
    ## generate output file(s) ##
    ##            END          ##
    #############################
    
  } # end of window i 
}



